import typing

import json

import numpy as np
import pandas as pd
import pydantic

from retrack.engine.parser import Parser
from retrack.engine.request_manager import RequestManager
from retrack.nodes.base import NodeKind, NodeMemoryType
from retrack.utils import constants


class Runner:
    def __init__(self, parser: Parser, name: str = None):
        self._parser = parser
        self._name = name
        self._internal_runners = {}
        self.reset()
        self._set_constants()
        self._set_input_columns()
        self._request_manager = RequestManager(self._parser.get_by_kind(NodeKind.INPUT))
        self._set_internal_runners()

    @classmethod
    def from_json(cls, data: typing.Union[str, dict], name: str = None, **kwargs):
        if isinstance(data, str) and data.endswith(".json"):
            if name is None:
                name = data
            data = json.loads(open(data).read())
        elif not isinstance(data, dict):
            raise ValueError("data must be a dict or a json file path")

        parser = Parser(data, **kwargs)
        return cls(parser, name=name)

    @property
    def parser(self) -> Parser:
        return self._parser

    @property
    def name(self) -> str:
        return self._name

    @property
    def request_manager(self) -> RequestManager:
        return self._request_manager

    @property
    def request_model(self) -> pydantic.BaseModel:
        return self._request_manager.model

    @property
    def states(self) -> pd.DataFrame:
        return self._states

    @property
    def filters(self) -> dict:
        return self._filters

    @property
    def constants(self) -> dict:
        return self._constants

    def _set_constants(self):
        constant_nodes = self.parser.get_by_memory_type(NodeMemoryType.CONSTANT)
        self._constants = {}
        for node in constant_nodes:
            for output_connector_name, _ in node.outputs:
                self._constants[f"{node.id}@{output_connector_name}"] = node.data.value

    def _set_internal_runners(self):
        for node_id in self.parser.indexes_by_name_map.get(
            constants.FLOW_NODE_NAME, []
        ):
            try:
                node_data = self.parser.get_by_id(node_id).data
                self._internal_runners[node_id] = Runner.from_json(
                    node_data.parsed_value(), name=node_data.name
                )
            except Exception as e:
                raise Exception(
                    f"Error setting internal runner for node {node_id}"
                ) from e

    @property
    def input_columns(self) -> dict:
        return self._input_columns

    def _set_input_columns(self):
        input_nodes = self._parser.get_by_kind(NodeKind.INPUT)
        self._input_columns = {
            f"{node.id}@{constants.INPUT_OUTPUT_VALUE_CONNECTOR_NAME}": node.data.name
            for node in input_nodes
        }

    def reset(self):
        self._states = None
        self._filters = {}

    def __set_output_connection_filters(
        self, node_id: str, filter: typing.Any, filter_by_connector=None
    ):
        if filter is not None:
            output_connections = self.parser.get_node_connections(
                node_id, is_input=False, filter_by_connector=filter_by_connector
            )
            for output_connection_id in output_connections:
                if self._filters.get(output_connection_id, None) is None:
                    self._filters[output_connection_id] = filter
                else:
                    self._filters[output_connection_id] = (
                        self._filters[output_connection_id] & filter
                    )

    def _create_initial_state_from_payload(
        self, payload_df: pd.DataFrame
    ) -> pd.DataFrame:
        """Create initial state from payload. This is the first step of the runner."""
        validated_payload = self.request_manager.validate(
            payload_df.reset_index(drop=True)
        )

        state_df = pd.DataFrame([])
        for node_id, input_name in self.input_columns.items():
            state_df[node_id] = validated_payload[input_name]

        state_df[constants.OUTPUT_REFERENCE_COLUMN] = np.nan
        state_df[constants.OUTPUT_MESSAGE_REFERENCE_COLUMN] = np.nan

        return state_df

    def __get_input_params(
        self, node_id: str, node_dict: dict, current_node_filter: pd.Series
    ) -> dict:
        input_params = {}

        for connector_name, connections in node_dict.get("inputs", {}).items():
            if connector_name.endswith(constants.NULL_SUFFIX):
                continue

            for connection in connections["connections"]:
                input_params[connector_name] = self.__get_state_data(
                    f"{connection['node']}@{connection['output']}", current_node_filter
                )

        if node_id in self._internal_runners:
            input_params["runner"] = self._internal_runners[node_id]

        return input_params

    def __set_state_data(
        self, column: str, value: typing.Any, filter_by: typing.Any = None
    ):
        if filter_by is None:
            self._states[column] = value
        else:
            self._states.loc[filter_by, column] = value

    def __get_state_data(self, column: str, filter_by: typing.Any = None):
        if column in self._constants:
            return self._constants[column]

        if filter_by is None:
            return self._states[column]

        return self._states.loc[filter_by, column]

    def __run_node(self, node_id: str):
        current_node_filter = self._filters.get(node_id, None)
        # if there is a filter, we need to set the children nodes to receive filtered data
        self.__set_output_connection_filters(node_id, current_node_filter)

        node = self.parser.get_by_id(node_id)

        if node.memory_type == NodeMemoryType.CONSTANT:
            return

        input_params = self.__get_input_params(
            node_id, node.model_dump(by_alias=True), current_node_filter
        )
        output = node.run(**input_params)

        for output_name, output_value in output.items():
            if (
                output_name == constants.OUTPUT_REFERENCE_COLUMN
                or output_name == constants.OUTPUT_MESSAGE_REFERENCE_COLUMN
            ):  # Setting output values
                self.__set_state_data(output_name, output_value, current_node_filter)
            elif output_name.endswith(constants.FILTER_SUFFIX):  # Setting filters
                self.__set_output_connection_filters(node_id, output_value, output_name)
            else:  # Setting node outputs to be used as inputs by other nodes
                self.__set_state_data(
                    f"{node_id}@{output_name}", output_value, current_node_filter
                )

    def execute(
        self,
        payload_df: pd.DataFrame,
        return_all_states: bool = False,
    ) -> pd.DataFrame:
        """Executes the flow with the given payload.

        Args:
            payload_df (pd.DataFrame): The payload to be used as input.
            return_all_states (bool, optional): If True, returns all states. Defaults to False.

        Returns:
            pd.DataFrame: The output of the flow.
        """
        if not isinstance(payload_df, pd.DataFrame):
            raise ValueError("payload_df must be a pandas.DataFrame")

        self.reset()
        self._states = self._create_initial_state_from_payload(payload_df)

        for node_id in self.parser.execution_order:
            try:
                self.__run_node(node_id)
            except Exception as e:
                raise Exception(
                    f"Error running node {node_id} in {self.name} with version {self.parser.version}"
                ) from e

            if self.states[constants.OUTPUT_REFERENCE_COLUMN].isna().sum() == 0:
                break

        if return_all_states:
            return self.states

        return self.states[
            [
                constants.OUTPUT_REFERENCE_COLUMN,
                constants.OUTPUT_MESSAGE_REFERENCE_COLUMN,
            ]
        ]



from retrack.utils.registry import Registry
from retrack.validators.base import BaseValidator
from retrack.validators.check_is_dag import CheckIsDAG
from retrack.validators.node_exists import NodeExistsValidator

_registry = Registry()


def registry() -> Registry:
    return _registry


def register(name: str, validator: BaseValidator) -> None:
    registry().register(name, validator)


register(
    "single_start_node_exists",
    NodeExistsValidator("start", min_quantity=1, max_quantity=1),
)
register("check_is_dag", CheckIsDAG())


__all__ = ["registry", "register", "BaseValidator"]



from retrack.engine.parser import Parser
from retrack.engine.runner import Runner
from retrack.nodes import registry as component_registry
from retrack.nodes.base import BaseNode, InputConnectionModel, OutputConnectionModel

__all__ = [
    "Parser",
    "Runner",
    "BaseNode",
    "InputConnectionModel",
    "OutputConnectionModel",
    "component_registry",
]



import typing

from retrack.nodes.dynamic.base import BaseDynamicNode
from retrack.nodes.dynamic.csv_table import csv_table_factory
from retrack.nodes.dynamic.flow import flow_factory
from retrack.utils.registry import Registry

_registry = Registry()


def registry() -> Registry:
    return _registry


def register(
    name: str,
    factory: typing.Callable[
        [typing.List[str], typing.List[str]], typing.Type[BaseDynamicNode]
    ],
) -> None:
    registry().register(name, factory)


register("CSVTableV0", csv_table_factory)
register("FlowV0", flow_factory)

__all__ = ["registry", "register", "BaseDynamicNode"]



import typing

import hashlib

from retrack import nodes, validators
from retrack.utils.registry import Registry
import json


class Parser:
    def __init__(
        self,
        graph_data: dict,
        component_registry: Registry = nodes.registry(),
        dynamic_registry: Registry = nodes.dynamic_registry(),
        validator_registry: Registry = validators.registry(),
    ):
        self.__graph_data = graph_data
        self._execution_order = None
        self.__components = {}
        self.__edges = None

        self._check_input_data(self.graph_data)

        self._set_components(component_registry, dynamic_registry)
        self._set_edges()

        self._validate_graph(validator_registry)

        self._set_indexes_by_name_map()
        self._set_indexes_by_kind_map()
        self._set_execution_order()
        self._set_indexes_by_memory_type_map()
        self._set_version()

    @property
    def graph_data(self) -> dict:
        return self.__graph_data

    @property
    def version(self) -> str:
        return self._version

    @staticmethod
    def _check_input_data(data: dict):
        if not isinstance(data, dict):
            raise TypeError(
                "Data must be a dictionary. Instead got: " + str(type(data))
            )
        if "nodes" not in data:
            raise ValueError("No nodes found in data")
        if not isinstance(data["nodes"], dict):
            raise TypeError(
                "BaseNodes must be a dictionary. Instead got: "
                + str(type(data["nodes"]))
            )

    @staticmethod
    def _check_node_name(node_name: str, node_id: str):
        if node_name is None:
            raise ValueError(f"BaseNode {node_id} has no name")
        if not isinstance(node_name, str):
            raise TypeError(f"BaseNode {node_id} name must be a string")

    @property
    def components(self) -> typing.Dict[str, nodes.BaseNode]:
        return self.__components

    def _set_components(self, component_registry: Registry, dynamic_registry: Registry):
        for node_id, node_metadata in self.graph_data["nodes"].items():
            if node_id in self.__components:
                raise ValueError(f"Duplicate node id: {node_id}")

            node_name = node_metadata.get("name", None)
            self._check_node_name(node_name, node_id)

            node_name = node_name.lower()

            node_factory = dynamic_registry.get(node_name)

            if node_factory is not None:
                validation_model = node_factory(**node_metadata)
            else:
                validation_model = component_registry.get(node_name)

            if validation_model is None:
                raise ValueError(f"Unknown node name: {node_name}")

            self.__components[node_id] = validation_model(**node_metadata)

    @property
    def edges(self) -> typing.List[typing.Tuple[str, str]]:
        return self.__edges

    def _set_edges(self):
        self.__edges = []

        for node_id, node in self.components.items():
            for _, output_connection in node.outputs:
                for c in output_connection.connections:
                    self.__edges.append((node_id, c.node))

    def _validate_graph(self, validator_registry: Registry):
        for validator_name, validator in validator_registry.data.items():
            if not validator.validate(graph_data=self.graph_data, edges=self.edges):
                raise ValueError(f"Invalid graph data: {validator_name}")

    def get_by_id(self, id_: str) -> nodes.BaseNode:
        return self.components.get(id_)

    @property
    def indexes_by_name_map(self) -> typing.Dict[str, typing.List[str]]:
        return self._indexes_by_name_map

    def _set_indexes_by_name_map(self):
        self._indexes_by_name_map = {}

        for node_id, node in self.components.items():
            node_name = node.__class__.__name__.lower()
            if node_name not in self._indexes_by_name_map:
                self._indexes_by_name_map[node_name] = []

            self._indexes_by_name_map[node_name].append(node_id)

    def get_by_name(self, name: str) -> typing.List[nodes.BaseNode]:
        name = name.lower()
        return [self.get_by_id(id_) for id_ in self.indexes_by_name_map.get(name, [])]

    @property
    def indexes_by_kind_map(self) -> typing.Dict[str, typing.List[str]]:
        return self._indexes_by_kind_map

    def _set_indexes_by_kind_map(self):
        self._indexes_by_kind_map = {}

        for node_id, node in self.components.items():
            if node.kind() not in self._indexes_by_kind_map:
                self._indexes_by_kind_map[node.kind()] = []

            self._indexes_by_kind_map[node.kind()].append(node_id)

    def get_by_kind(self, kind: str) -> typing.List[nodes.BaseNode]:
        return [self.get_by_id(id_) for id_ in self.indexes_by_kind_map.get(kind, [])]

    @property
    def indexes_by_memory_type_map(self) -> typing.Dict[str, typing.List[str]]:
        return self._indexes_by_memory_type_map

    def _set_indexes_by_memory_type_map(self):
        self._indexes_by_memory_type_map = {}

        for node_id, node in self.components.items():
            memory_type = node.memory_type()
            if memory_type not in self.indexes_by_memory_type_map:
                self._indexes_by_memory_type_map[memory_type] = []

            self._indexes_by_memory_type_map[memory_type].append(node_id)

    def get_by_memory_type(self, memory_type: str) -> typing.List[nodes.BaseNode]:
        return [
            self.get_by_id(id_)
            for id_ in self.indexes_by_memory_type_map.get(memory_type, [])
        ]

    @property
    def execution_order(self) -> typing.List[str]:
        return self._execution_order

    def _set_execution_order(self):
        start_nodes = self.get_by_name("start")

        self._execution_order = self._walk(start_nodes[0].id, [])

    def get_node_connections(
        self, node_id: str, is_input: bool = True, filter_by_connector=None
    ):
        node_dict = self.get_by_id(node_id).model_dump(by_alias=True)

        connectors = node_dict.get("inputs" if is_input else "outputs", {})
        result = []

        for connector_name, value in connectors.items():
            if (
                filter_by_connector is not None
                and connector_name != filter_by_connector
            ):
                continue

            for connection in value["connections"]:
                result.append(connection["node"])
        return result

    def _walk(self, actual_id: str, skiped_ids: list):
        skiped_ids.append(actual_id)

        output_ids = self.get_node_connections(actual_id, is_input=False)

        for next_id in output_ids:
            if next_id not in skiped_ids:
                next_node_input_ids = self.get_node_connections(next_id, is_input=True)
                run_next = True
                for next_node_input_id in next_node_input_ids:
                    if next_node_input_id not in skiped_ids:
                        run_next = False
                        break

                if run_next:
                    self._walk(next_id, skiped_ids)

        return skiped_ids

    def _set_version(self):
        self._version = self.graph_data.get("version", None)

        graph_json_content = (
            json.dumps(self.graph_data["nodes"])
            .replace(": ", ":")
            .replace(", ", ",")
            .encode("utf-8")
        )
        calculated_hash = hashlib.sha256(graph_json_content).hexdigest()[:10]

        if self.version is None:
            self._version = f"{calculated_hash}.dynamic"
        else:
            file_version_hash = self.version.split(".")[0]

            if file_version_hash != calculated_hash:
                raise ValueError(
                    f"Invalid version. Graph data has changed and the hash is different: {calculated_hash} != {file_version_hash}"
                )
